"""Test if micom simulation works."""

import os.path as path
import pandas as pd
import qiime2 as q2
import q2_micom as q2m
from q2_micom.tests import this_dir

this_dir = q2m.tests.this_dir

medium = q2.Artifact.load(path.join(this_dir, "data", "medium.qza")).view(
    pd.DataFrame
)
models = q2.Artifact.load(path.join(this_dir, "data", "build.qza"))

res = q2m.grow(
    models.view(q2m._formats_and_types.CommunityModelDirectory),
    medium
)


def test_growth_rates():
    gcs = res.growth_rates.view(pd.DataFrame)
    assert "growth_rate" in gcs.columns
    assert "sample_id" in gcs.columns
    assert all(gcs.growth_rate > 1e-6)


def test_exchanges():
    ex = res.exchange_fluxes.view(pd.DataFrame)
    assert "reaction" in ex.columns
    assert ex.reaction.str.startswith("EX_").all()


def test_feasible_exchanges():
    ex = res.exchange_fluxes.view(pd.DataFrame)
    ex = ex[(ex.taxon == "medium") & (ex.direction == "import")]
    ex["bound"] = medium.loc[ex.reaction, "flux"].values
    ex = ex.dropna()
    assert all(ex.flux < ex.bound + 1e-6)



"""Global imports for q2-micom."""

from q2_micom._build import build
from q2_micom._db import db
from q2_micom._formats_and_types import MicomResultsData
from q2_micom._growth import grow
from q2_micom._medium import minimal_medium
from q2_micom._tradeoff import tradeoff
from q2_micom._viz import (
    plot_growth,
    exchanges_per_sample,
    exchanges_per_taxon,
    plot_tradeoff,
    fit_phenotype,
)

__version__ = "0.10.0"


def read_results(path):
    """Read the results from a MICOM simulation.

    Parameters:
    -----------
    path : str
        The path to a MicomResults artifact.

    Returns:
    --------
    MicomResultsData
        A named tuple with the following attributes:

        growth_rates : pd.DataFrame
        The growth rates for each taxon and sample.

        exchange_fluxes : pd.DataFrame
        The exchange fluxes for each metabolite, sample and taxon. Fluxes
        that denote trasnport from and into the environment are denoted with
        the taxon `medium.
    """
    from qiime2 import Artifact
    art = Artifact.load(path)
    return art.view(MicomResultsData)


__all__ = [
    "db",
    "build",
    "minimal_medium",
    "grow",
    "tradeoff",
    "plot_growth",
    "exchanges_per_sample",
    "exchanges_per_taxon",
    "plot_tradeoff",
    "fit_phenotype",
    "read_results",
]



"""Test if visualizations work."""

from micom.workflows.core import GrowthResults
import os.path as path
import pandas as pd
import pytest
import qiime2 as q2
import q2_micom as q2m
from tempfile import TemporaryDirectory

this_dir = q2m.tests.this_dir

results = q2.Artifact.load(path.join(this_dir, "data", "growth.qza"))


def test_growth_plots():
    r = results.view(GrowthResults)
    with TemporaryDirectory(prefix="q2-micom-") as d:
        q2m.plot_growth(str(d), r)
        assert q2m.tests.check_viz(str(d))


def test_exchanges_per_sample():
    r = results.view(GrowthResults)
    with TemporaryDirectory(prefix="q2-micom-") as d:
        q2m.exchanges_per_sample(str(d), r)
        assert q2m.tests.check_viz(str(d))


def test_exchanges_per_taxon():
    r = results.view(GrowthResults)
    with TemporaryDirectory(prefix="q2-micom-") as d:
        q2m.exchanges_per_taxon(str(d), r)
        assert q2m.tests.check_viz(str(d))


def test_plot_tradeoff():
    tradeoff = q2.Artifact.load(path.join(this_dir, "data", "tradeoff.qza"))
    t = tradeoff.view(pd.DataFrame)
    with TemporaryDirectory(prefix="q2-micom-") as d:
        q2m.plot_tradeoff(str(d), t)
        assert q2m.tests.check_viz(str(d))


def test_fit_phenotype():
    large = q2.Artifact.load(
        path.join(this_dir, "data", "growth.qza"))
    r = large.view(GrowthResults)
    mcol = q2.Metadata.load(
        path.join(this_dir, "data", "metadata.tsv")
    ).get_column("status")
    with TemporaryDirectory(prefix="q2-micom-") as d:
        q2m.fit_phenotype(str(d), r, mcol, min_coef=0.001)
        assert q2m.tests.check_viz(str(d))
    with TemporaryDirectory(prefix="q2-micom-") as d:
        q2m.fit_phenotype(str(d), r, mcol, flux_type="import",
                          min_coef=0.0)
        assert q2m.tests.check_viz(str(d))



"""Constructs a minimal medium for a set of community models."""

from cobra.util.solver import OptimizationError
from micom import load_pickle
import micom.media as mm
import micom.workflows as mw
import pandas as pd
from q2_micom._formats_and_types import CommunityModelDirectory


def minimal_medium(
    models: CommunityModelDirectory, min_growth: float = 0.1, threads: int = 1
) -> pd.DataFrame:
    """Calculate the minimal medium for a set of community models."""
    manifest = models.manifest.view(pd.DataFrame)
    model_folder = (
        str(models.model_files.path_maker(model_id="blub"))
        .replace("blub.pickle", "")
    )
    medium = mw.minimal_media(manifest, model_folder, True,
                              min_growth, threads)
    return medium



"""Test if minimal medium works."""

import biom
import micom
import os.path as path
import pandas as pd
import qiime2 as q2
import q2_micom as q2m

this_dir = q2m.tests.this_dir

models = q2.Artifact.load(path.join(this_dir, "data", "build.qza"))

medium_high = q2m._medium.minimal_medium(
    models.view(q2m._formats_and_types.CommunityModelDirectory),
    0.5
)

medium_low = q2m._medium.minimal_medium(
    models.view(q2m._formats_and_types.CommunityModelDirectory),
    0.1
)


def test_medium():
    assert "reaction" in medium_high.columns
    assert "flux" in medium_high.columns
    assert all(medium_high.flux > 1e-6)


def test_sane_medium():
    assert medium_high.flux.sum() > medium_low.flux.sum()



"""Build a database of organism metabolic models."""

from micom.workflows import build_database
import os
from qiime2 import Metadata
from q2_micom._formats_and_types import JSONDirectory


def db(
    meta: Metadata, rank: str = "genus", threads: int = 1
) -> JSONDirectory:
    """Create a model database from a set of SBML files."""
    meta = meta.to_dataframe()
    json_dir = JSONDirectory()
    path = str(json_dir.json_files.path_maker(model_id="dummy"))
    path = os.path.dirname(path)
    meta = build_database(meta, path, rank, threads, compress=False)
    os.rename(
        os.path.join(path, "manifest.csv"),
        json_dir.manifest.path_maker())
    return json_dir



"""Build Community models."""

import biom
import os
import micom.workflows as mw
from micom.taxonomy import build_from_qiime
import pandas as pd
from q2_micom._formats_and_types import (
    JSONDirectory,
    CommunityModelDirectory,
)


RANKS = [
    "kingdom",
    "phylum",
    "class",
    "order",
    "family",
    "genus",
    "species",
    "strain",
]


def build_spec(
    abundance: biom.Table,
    taxonomy: pd.Series,
    models: JSONDirectory,
    cutoff: float,
    strict: bool
) -> pd.DataFrame:
    """Build the specification for the community models."""
    model_files = models.manifest.view(pd.DataFrame).rename(columns={"id": "model_id"})
    rank = model_files.summary_rank[0]
    rank_index = RANKS.index(rank)
    if strict:
        ranks = RANKS[0:rank_index]
    else:
        ranks = [rank]
    model_files["file"] = model_files[rank].apply(
        lambda i: str(models.json_files.path_maker(model_id=i))
    )

    tax = build_from_qiime(abundance, taxonomy, collapse_on=ranks)
    micom_taxonomy = pd.merge(model_files, tax, on=ranks)
    micom_taxonomy = micom_taxonomy[micom_taxonomy.relative > cutoff]
    stats = micom_taxonomy.sample_id.value_counts().describe()
    print("Merged with the database using ranks: %s" % ", ".join(ranks))
    print(
        "Each community model contains %d-%d taxa (average %d+-%d)." % (
            stats["min"], stats["max"],
            round(stats["mean"]), round(stats["std"]))
    )
    return micom_taxonomy


def build(
    abundance: biom.Table,
    taxonomy: pd.Series,
    models: JSONDirectory,
    threads: int = 1,
    cutoff: float = 0.0001,
    strict: bool = False,
) -> CommunityModelDirectory:
    """Build the community models."""
    tax = build_spec(abundance, taxonomy, models, cutoff, strict)
    out = CommunityModelDirectory()
    out_folder = (
        str(out.model_files.path_maker(model_id="test"))
        .replace("test.pickle", "")
    )
    model_folder = (
        str(models.json_files.path_maker(model_id="test"))
        .replace("test.json", "")
    )
    mw.build(tax, model_folder, out_folder, cutoff, threads)
    os.rename(os.path.join(out_folder, "manifest.csv"),
              out.manifest.path_maker())
    return out



"""Helpers to run cooperative tradeoff with various tradeoff values."""

import micom.workflows as mw
import numpy as np
import pandas as pd
from q2_micom._formats_and_types import (
    CommunityModelDirectory,
)


def tradeoff(
    models: CommunityModelDirectory,
    medium: pd.DataFrame,
    tradeoff_min: float = 0.1,
    tradeoff_max: float = 1.0,
    step: float = 0.1,
    threads: int = 1,
) -> pd.DataFrame:
    """Run tradeoff analysis."""
    manifest = models.manifest.view(pd.DataFrame)
    model_folder = (
        str(models.model_files.path_maker(model_id="blub"))
        .replace("blub.pickle", "")
    )
    if tradeoff_min >= tradeoff_max:
        raise ValueError(
            "`tradeoff_min` must be smaller than `tradeoff_max` :("
        )
    tradeoffs = np.arange(tradeoff_min, tradeoff_max + 1e-6, step)
    results = mw.tradeoff(manifest, model_folder, medium, tradeoffs, threads)
    return results



"""Performs a growth simulation."""

import micom.workflows as mw
import pandas as pd
from q2_micom._formats_and_types import (
    MicomResultsDirectory,
    CommunityModelDirectory,
)


def grow(
    models: CommunityModelDirectory,
    medium: pd.DataFrame,
    tradeoff: float = 0.5,
    threads: int = 1,
) -> MicomResultsDirectory:
    """Simulate growth for a set of community models."""
    out = MicomResultsDirectory()
    model_folder = (
        str(models.model_files.path_maker(model_id="blub"))
        .replace("blub.pickle", "")
    )
    manifest = models.manifest.view(pd.DataFrame)
    growth, exchanges, annotations = mw.grow(
        manifest, model_folder, medium, tradeoff, threads)
    growth.to_csv(out.growth_rates.path_maker(), index=False)
    annotations.to_csv(out.annotations.path_maker(), index=False)
    exchanges[pd.notna(exchanges.flux)].to_csv(
        out.exchange_fluxes.path_maker(), index=False
    )
    return out



"""Formats and types for metabolic modeling."""

from collections import namedtuple
import pandas as pd
from qiime2.plugin import SemanticType
import qiime2.plugin.model as model

REQ_FIELDS = pd.Series(["file", "kingdom", "phylum", "class", "order",
                        "family", "genus", "species"])


class SBMLFormat(model.TextFileFormat):
    """Represents an SBML file."""
    def _check_n_lines(self, n):
        """Crudely check if the file is SBML."""
        with open(str(self), mode="r") as xml_file:
            lines = "".join(xml_file.readlines(n))
        return ".xml" in str(self).lower() and "<sbml" in lines.lower()

    def _validate_(self, level):
        record_map = {'min': 5, 'max': 100}
        return self._check_n_lines(record_map[level])


class JSONFormat(model.TextFileFormat):
    """Represents a JSON file."""
    def _check_n_lines(self, n):
        """Crudely check if the file is SBML."""
        with open(str(self), mode="r") as xml_file:
            lines = "".join(xml_file.readlines(n))
        return ".xml" in str(self).lower() and "metabolites" in lines.lower()

    def _validate_(self, level):
        record_map = {'min': 5, 'max': 100}
        return self._check_n_lines(record_map[level])


class ModelManifest(model.TextFileFormat):
    """Represents an SBML file."""

    def _validate_(self, level):
        header = open(str(self), mode="r").readline().split(",")
        return REQ_FIELDS.isin(header).all()


class SBMLDirectory(model.DirectoryFormat):
    manifest = model.File("manifest.csv", format=ModelManifest)
    sbml_files = model.FileCollection(r".+\.xml", format=SBMLFormat)

    @sbml_files.set_path_maker
    def sbml_path_maker(self, model_id):
        return "%s.xml" % model_id


class JSONDirectory(model.DirectoryFormat):
    manifest = model.File("manifest.csv", format=ModelManifest)
    json_files = model.FileCollection(r".+\.json", format=JSONFormat)

    @json_files.set_path_maker
    def sbml_path_maker(self, model_id):
        return "%s.json" % model_id


class CommunityModelFormat(model.BinaryFileFormat):
    """Represents a pickled community model."""
    def _validate_(self, level):
        return str(self).lower().endswith(".pickle")


class CommunityModelManifest(model.TextFileFormat):
    """Represents a manifets for community models."""

    def _validate_(self, level):
        header = open(str(self), mode="r").readline().split(",")
        return "sample_id" in header


class CommunityModelDirectory(model.DirectoryFormat):
    manifest = model.File("manifest.csv", format=CommunityModelManifest)
    model_files = model.FileCollection(r".+\.pickle",
                                       format=CommunityModelFormat)

    @model_files.set_path_maker
    def model_path_maker(self, model_id):
        return "%s.pickle" % model_id


class GrowthRates(model.TextFileFormat):
    def _validate_(self, level):
        header = open(str(self), mode="r").readline().split(",")
        return header == ["sample_id", "taxon", "reactions", "metabolites",
                          "abundance", "tradeoff", "growth_rate"]


class Fluxes(model.TextFileFormat):
    def _validate_(self, level):
        return str(self).endswith(".csv")


class Annotations(model.TextFileFormat):
    def _validate_(self, level):
        header = open(str(self), mode="r").readline().split(",")
        return header == ["reaction", "metabolite", "description"]


class MicomResultsDirectory(model.DirectoryFormat):
    growth_rates = model.File("growth_rates.csv", format=GrowthRates)
    exchange_fluxes = model.File("exchange_fluxes.csv",
                                 format=Fluxes)
    annotations = model.File("annotations.csv", format=Annotations)


class MicomMediumFile(model.TextFileFormat):
    def _validate_(self, level):
        header = open(str(self), mode="r").readline().split(",")
        return header == ["metabolite", "reaction", "flux"]


MicomMediumDirectory = model.SingleFileDirectoryFormat(
    "MicomMediumDirectory", "medium.csv", MicomMediumFile)

TradeoffResultsDirectory = model.SingleFileDirectoryFormat(
    "TradeoffResultsDirectory", "tradeoff.csv", GrowthRates)

SBML = SemanticType("SBML")
Pickle = SemanticType("Pickle")
JSON = SemanticType("JSON")

MetabolicModels = SemanticType(
    "MetabolicModels",
    field_names="format",
    field_members={"format": (SBML, JSON)}
)

CommunityModels = SemanticType(
    "CommunityModels",
    field_names="format",
    field_members={"format": (Pickle,)}
)

MicomResults = SemanticType("MicomResults")
TradeoffResults = SemanticType("TradeoffResults")

Global = SemanticType("Global")
PerSample = SemanticType("PerSample")
MicomMedium = SemanticType(
    "MicomMedium",
    field_names="type",
    field_members={"type": (Global, PerSample)})


MicomResultsData = namedtuple(
    "MicomResultsData", ["growth_rates", "exchange_fluxes"])



"""Plugin setup."""

import importlib
from qiime2.plugin import (
    Plugin,
    Str,
    Choices,
    Int,
    Bool,
    Range,
    Float,
    Metadata,
    MetadataColumn,
    Categorical,
    Numeric,
    Citations,
)

import q2_micom
from q2_micom._formats_and_types import (
    SBML,
    JSON,
    Pickle,
    SBMLFormat,
    SBMLDirectory,
    JSONFormat,
    JSONDirectory,
    CommunityModelFormat,
    CommunityModelManifest,
    CommunityModelDirectory,
    GrowthRates,
    Fluxes,
    MicomResultsDirectory,
    MicomMediumFile,
    MicomMediumDirectory,
    MetabolicModels,
    CommunityModels,
    MicomResults,
    MicomMedium,
    Global,
    PerSample,
    TradeoffResults,
    TradeoffResultsDirectory,
    REQ_FIELDS,
)
from q2_types.feature_data import FeatureData, Taxonomy
from q2_types.feature_table import FeatureTable, RelativeFrequency, Frequency

citations = Citations.load("citations.bib", package="q2_micom")


plugin = Plugin(
    name="micom",
    version=q2_micom.__version__,
    website="https://github.com/micom-dev/q2-micom",
    package="q2_micom",
    description=(""),
    short_description="Plugin for metabolic modeling of "
    "microbial communities.",
)

plugin.register_formats(
    SBMLFormat,
    SBMLDirectory,
    JSONFormat,
    JSONDirectory,
    CommunityModelFormat,
    CommunityModelManifest,
    CommunityModelDirectory,
    GrowthRates,
    Fluxes,
    MicomResultsDirectory,
    MicomMediumFile,
    MicomMediumDirectory,
    TradeoffResultsDirectory,
)
plugin.register_semantic_types(
    MetabolicModels, CommunityModels, MicomResults, MicomMedium
)
plugin.register_semantic_type_to_format(MetabolicModels[SBML], SBMLDirectory)
plugin.register_semantic_type_to_format(MetabolicModels[JSON], JSONDirectory)
plugin.register_semantic_type_to_format(
    CommunityModels[Pickle], CommunityModelDirectory
)
plugin.register_semantic_type_to_format(MicomResults, MicomResultsDirectory)
plugin.register_semantic_type_to_format(
    TradeoffResults, TradeoffResultsDirectory
)
plugin.register_semantic_type_to_format(
    MicomMedium[Global], MicomMediumDirectory
)
plugin.register_semantic_type_to_format(
    MicomMedium[PerSample], MicomMediumDirectory
)

plugin.methods.register_function(
    function=q2_micom.db,
    inputs={},
    parameters={
        "meta": Metadata,
        "rank": Str % Choices(q2_micom._build.RANKS),
        "threads": Int % Range(1, None),
    },
    outputs=[("metabolic_models", MetabolicModels[JSON])],
    input_descriptions={},
    parameter_descriptions={
        "meta": (
            "Metadata for the individual metabolic models in `folder`. "
            "Must contain the the following columns: %s."
            % ", ".join(REQ_FIELDS)
        ),
        "rank": "The phylogenetic rank at which to summarize taxa.",
        "threads": "The number of threads to use when constructing models.",
    },
    output_descriptions={"metabolic_models": "The metabolic model DB."},
    name="Build a metabolic model database.",
    description=(
        "Constructs pan-genome models summarized to the specified rank "
        "and bundles the models to be used by MICOM. "
        "The chosen rank has to be the same you want as when building your "
        "community models. "
        "So you may not build genus-level community models with a species "
        "level database. "
        "You will only need to run this function if you want to build a "
        "custom DB. For many use cases downloading the prebuilt AGORA DB "
        "with the the preferred rank should be sufficient."
    ),
    citations=[
        citations["agora"],
        citations["agora_reply"],
        citations["micom"],
    ],
)

plugin.methods.register_function(
    function=q2_micom.build,
    inputs={
        "abundance": FeatureTable[Frequency | RelativeFrequency],
        "taxonomy": FeatureData[Taxonomy],
        "models": MetabolicModels[JSON],
    },
    parameters={
        "threads": Int % Range(1, None),
        "cutoff": Float % Range(0.0, 1.0),
        "strict": Bool
    },
    outputs=[("community_models", CommunityModels[Pickle])],
    input_descriptions={
        "abundance": (
            "The feature table containing the samples over which beta "
            "diversity should be computed."
        ),
        "taxonomy": "The taxonomy assignments for the ASVs in the table.",
        "models": "The single taxon model database to use.",
    },
    parameter_descriptions={
        "threads": "The number of threads to use when constructing models.",
        "cutoff": "Taxa with a relative abundance smaller than this will "
        "be dropped.",
        "strict": "If true will collapse and match on all taxa ranks up to the "
                  "specified rank (so on all higher ranks as well). If false "
                  "(default) will match only on single taxa rank specified before. "
                  "If using the strict option make sure ranks are named the same as in "
                  "the used database."
    },
    output_descriptions={"community_models": "The community models."},
    name="Build community models.",
    description=(
        "Builds the metabolic community models for a " "set of samples."
    ),
    citations=[citations["micom"]],
)

plugin.methods.register_function(
    function=q2_micom.minimal_medium,
    inputs={"models": CommunityModels[Pickle]},
    parameters={
        "min_growth": Float % Range(0.0, None, inclusive_start=False),
        "threads": Int % Range(1, None),
    },
    outputs=[("medium", MicomMedium[Global])],
    input_descriptions={
        "models": (
            "A collection of metabolic community models. "
            "This should contain on model for each sample."
        ),
    },
    parameter_descriptions={
        "min_growth": (
            "The minimum achievable growth rate for each taxon. "
            "The returned growth medium enables all taxa to growth "
            "simultaneously with at least this rate."
        ),
        "threads": "The number of threads to use when simulating.",
    },
    output_descriptions={"medium": "The resulting growth medium."},
    name="Obtain a minimal growth medium for models.",
    description=(
        "Obtains a minimal growth medium for the community models. "
        "Please note that this medium does not have any biological "
        "feasibility. If you have any knowledge about metabolites present "
        "in the environment we recommend you construct the medium by hand."
    ),
    citations=[citations["micom"]],
)

plugin.methods.register_function(
    function=q2_micom.grow,
    inputs={
        "models": CommunityModels[Pickle],
        "medium": MicomMedium[Global | PerSample],
    },
    parameters={
        "tradeoff": Float
        % Range(0.0, 1.0, inclusive_start=False, inclusive_end=True),
        "threads": Int % Range(1, None),
    },
    outputs=[("results", MicomResults)],
    input_descriptions={
        "models": (
            "A collection of metabolic community models. "
            "This should contain on model for each sample."
        ),
        "medium": "The growth medium to use.",
    },
    parameter_descriptions={
        "tradeoff": (
            "The tradeoff parameter. This describes the balance "
            "between maximizing biomass production of the entire "
            "community and biomass production of individual taxa "
            '(ergo "egoistic" growth). A value of 1.0 would yield '
            "the best biomass production across the community but "
            "will only allow a few taxa to grow. Smaller values will "
            "allow more taxa to grow but will sacrifice overall "
            "biomass. A value of 0.5 (the default) has been shown to "
            "best reproduce growth rates in the human gut."
        ),
        "threads": "The number of threads to use when simulating.",
    },
    output_descriptions={
        "results": "The resulting taxa-level growth rates and metabolic "
        "exchange fluxes."
    },
    name="Simulate growth for community models.",
    description=(
        "Simulates growth for a set of samples. Note that those are "
        'sample-specific or "personalized" simulations, so each taxa'
        "may have different growth rates and metabolite usahe in each sample."
    ),
    citations=[citations["micom"]],
)

plugin.methods.register_function(
    function=q2_micom.tradeoff,
    inputs={
        "models": CommunityModels[Pickle],
        "medium": MicomMedium[Global | PerSample],
    },
    parameters={
        "tradeoff_min": Float % Range(0.0, 1.0, inclusive_start=False),
        "tradeoff_max": Float % Range(0.0, 1.0, inclusive_end=True),
        "step": Float % Range(0.0, 1.0),
        "threads": Int,
    },
    outputs=[("results", TradeoffResults)],
    input_descriptions={
        "models": (
            "A collection of metabolic community models. "
            "This should contain on model for each sample."
        ),
        "medium": "The growth medium to use.",
    },
    parameter_descriptions={
        "tradeoff_min": "The minimum tradeoff parameter to test. This should "
        "be larger than 0.0 and smaller than 1.0.",
        "tradeoff_max": "The maximum tradeoff parameter to test. This should "
        "be larger than 0.0 and smaller than 1.0 and also be"
        "larger than `tradeoff_min`.",
        "step": "The tradeoff value step size to use.",
        "threads": "The number of threads to use when simulating.",
    },
    output_descriptions={
        "results": "The resulting taxa-level growth rates for varying "
        "tradeoff values."
    },
    name="Test a variety of tradeoff values.",
    description=(
        "Simulates growth for a set of samples while varying the tradeoff "
        "between community and taxon biomass production. "
        "This can be used to characterize a good tradeoff value for a "
        "specific set of samples. Our study suggested that a good tradeoff "
        "value is the largest value that allows the majority of taxa in the "
        "sample to grow."
    ),
    citations=[citations["micom"]],
)

plugin.visualizers.register_function(
    function=q2_micom.plot_growth,
    inputs={"results": MicomResults},
    parameters={},
    input_descriptions={
        "results": (
            "A set of MICOM analysis results. "
            "Contains predicted groath rates and exchange fluxes."
        )
    },
    parameter_descriptions={},
    name="Plot taxa growth rates.",
    description=(
        "Plot predicted growth rates for each taxon in each sample. "
        "Only points with growing taxa are shown (growth rate sufficiently "
        "larger than zero)."
    ),
    citations=[citations["micom"]],
)

plugin.visualizers.register_function(
    function=q2_micom.exchanges_per_sample,
    inputs={"results": MicomResults},
    parameters={
        "direction": Str % Choices("import", "export"),
        "cluster": Bool,
    },
    input_descriptions={
        "results": (
            "A set of MICOM analysis results. "
            "Contains predicted groath rates and exchange fluxes."
        )
    },
    parameter_descriptions={
        "direction": "The direction of the flux.",
        "cluster": "Whether to perform clutering on samples and reactions.",
    },
    name="Plot gloabl exchange rates.",
    description=(
        "Plot predicted global exchange fluxes for each sample. "
        "When plotting imports this corresponds to the consumption "
        "fluxes for each metabolite that is available to the community. "
        "When plotting export this corresponds to the production fluxes "
        "for each metabolite."
    ),
    citations=[citations["micom"]],
)


plugin.visualizers.register_function(
    function=q2_micom.exchanges_per_taxon,
    inputs={"results": MicomResults},
    parameters={
        "direction": Str % Choices("import", "export"),
        "perplexity": Int % Range(2, None),
    },
    input_descriptions={
        "results": (
            "A set of MICOM analysis results. "
            "Contains predicted growth rates and exchange fluxes."
        )
    },
    parameter_descriptions={
        "direction": "The direction of the flux.",
        "perplexity": "TSNE parameter. Relates to the number of neighbors used to "
        "calculate distances. Smaller values preserve more local "
        "structure and larger values preserve more global structure.",
    },
    name="Plot niche overlap.",
    description=(
        "Plot growth or production niches. "
        "The entire set of import or export fluxes for each taxon in each "
        "sample is reduced onto a single point on a 2D plane."
        "Taxa that are close to each other either consume similar metabolites "
        " (imports) or produce similar metabolites (exports)."
    ),
    citations=[citations["micom"]],
)

plugin.visualizers.register_function(
    function=q2_micom.plot_tradeoff,
    inputs={"results": TradeoffResults},
    parameters={},
    input_descriptions={
        "results": (
            "A set of MICOM tradeoff analysis results. "
            "Contains predicted growth rates for each tested tradeoff."
        )
    },
    parameter_descriptions={},
    name="Plot tradeoff results.",
    description=(
        "Plot predicted growth rate distributions for each tradeoff as "
        "well as the fraction of growing taxa in each sample and tradeoff "
        "value. For a good tradeoff value one usually tries to find the "
        "largest tradeoff value that still aloows most taxa to grow."
    ),
    citations=[citations["micom"]],
)

plugin.visualizers.register_function(
    function=q2_micom.fit_phenotype,
    inputs={"results": MicomResults},
    parameters={
        "metadata": MetadataColumn[Categorical | Numeric],
        "variable_type": Str % Choices("binary", "continuous"),
        "flux_type": Str % Choices("import", "production"),
        "min_coef": Float % Range(0, None),
    },
    input_descriptions={
        "results": (
            "A set of MICOM analysis results. "
            "Contains predicted growth rates and exchange fluxes."
        ),
    },
    parameter_descriptions={
        "metadata": "The metadata variable to use.",
        "variable_type": "The type of the phenotype variable.",
        "flux_type": "Which fluxes to use.",
        "min_coef": (
            "Only coefficient with absolute values larger than this "
            "will be shown."
        ),
    },
    name="Test for differential production",
    description=(
        "Test for overall metabolite production differences "
        "between two groups."
    ),
    citations=[citations["micom"]],
)

importlib.import_module("q2_micom._transform")



"""Visualizers for MICOM results."""

from micom.workflows.core import GrowthResults
import micom.viz as viz
from os.path import join
import pandas as pd
from qiime2 import MetadataColumn


def plot_growth(output_dir: str, results: GrowthResults) -> None:
    """Plot the taxa growth rates."""
    viz.plot_growth(
        results,
        join(output_dir, "index.html"),
    )


def exchanges_per_sample(
    output_dir: str,
    results: GrowthResults,
    direction: str = "import",
    cluster: bool = True,
) -> None:
    """Plot the exchange fluxes."""
    viz.plot_exchanges_per_sample(
        results,
        join(output_dir, "index.html"),
        direction,
        cluster
    )


def exchanges_per_taxon(
    output_dir: str,
    results: GrowthResults,
    direction: str = "import",
    perplexity: int = 20,
) -> None:
    """Plot the exchange fluxes."""
    viz.plot_exchanges_per_taxon(
        results,
        join(output_dir, "index.html"),
        direction,
        perplexity=perplexity,
    )


def plot_tradeoff(output_dir: str, results: pd.DataFrame) -> None:
    """Plot the taxa growth rates."""
    viz.plot_tradeoff(
        results,
        join(output_dir, "index.html")
    )


def fit_phenotype(
    output_dir: str,
    results: GrowthResults,
    metadata: MetadataColumn,
    variable_type: str = "binary",
    flux_type: str = "production",
    min_coef: float = 0.001,
):
    """Test for differential metabolite production."""
    meta = metadata.to_series()
    viz.plot_fit(
        results,
        meta,
        variable_type,
        meta.name,
        join(output_dir, "index.html"),
        flux_type,
        min_coef,
        atol=1e-6,
    )



"""Transformers for MICOM types."""

from micom.workflows.core import GrowthResults
import pandas as pd
from q2_micom.plugin_setup import plugin
import q2_micom._formats_and_types as ft


@plugin.register_transformer
def _1(data: pd.DataFrame) -> ft.MicomMediumFile:
    mm = ft.MicomMediumFile()
    data.to_csv(str(mm), index=False)
    return mm


@plugin.register_transformer
def _2(mm: ft.MicomMediumFile) -> pd.DataFrame:
    return pd.read_csv(str(mm), index_col=False)


@plugin.register_transformer
def _3(data: pd.DataFrame) -> ft.ModelManifest:
    sbm = ft.SBMLManifest()
    data.to_csv(str(sbm), index=False)
    return sbm


@plugin.register_transformer
def _4(sbm: ft.ModelManifest) -> pd.DataFrame:
    return pd.read_csv(str(sbm), index_col=False)


@plugin.register_transformer
def _5(data: pd.DataFrame) -> ft.CommunityModelManifest:
    cmm = ft.CommunityModelManifest()
    data.to_csv(str(cmm), index=False)
    return cmm


@plugin.register_transformer
def _6(cmm: ft.CommunityModelManifest) -> pd.DataFrame:
    return pd.read_csv(str(cmm), index_col=False)


@plugin.register_transformer
def _7(data: pd.DataFrame) -> ft.GrowthRates:
    gr = ft.GrowthRates()
    data.to_csv(str(gr), index=False)
    return gr


@plugin.register_transformer
def _8(gr: ft.GrowthRates) -> pd.DataFrame:
    return pd.read_csv(str(gr), index_col=False)


@plugin.register_transformer
def _9(data: pd.DataFrame) -> ft.Fluxes:
    ef = ft.Fluxes()
    data.to_csv(str(ef), index=False)
    return ef


@plugin.register_transformer
def _10(ef: ft.Fluxes) -> pd.DataFrame:
    return pd.read_csv(str(ef), index_col=False)


@plugin.register_transformer
def _11(res: ft.MicomResultsDirectory) -> ft.MicomResultsData:
    return ft.MicomResultsData(
        exchange_fluxes=pd.read_csv(str(res.exchange_fluxes.path_maker()),
                                    index_col=False),
        growth_rates=pd.read_csv(
            str(res.growth_rates.path_maker()), index_col=False)
    )


@plugin.register_transformer
def _12(data: pd.DataFrame) -> ft.Annotations:
    gr = ft.Annotations()
    data.to_csv(str(gr), index=False)
    return gr


@plugin.register_transformer
def _13(an: ft.Annotations) -> pd.DataFrame:
    return pd.read_csv(str(an), index_col=False)


@plugin.register_transformer
def _14(rd: ft.MicomResultsDirectory) -> GrowthResults:
    gr = GrowthResults(
        rd.growth_rates.view(pd.DataFrame),
        rd.exchange_fluxes.view(pd.DataFrame),
        rd.annotations.view(pd.DataFrame)
    )
    return gr


@plugin.register_transformer
def _15(gr: GrowthResults) -> ft.MicomResultsDirectory:
    rd = ft.MicomResultsDirectory()
    gr["growth_rates"].to_csv(rd.growth_rates.path_maker(), index=False)
    gr["annotations"].to_csv(rd.annotations.path_maker(), index=False)
    gr["exchanges"].to_csv(rd.exchange_fluxes.path_maker(), index=False)
    return rd




"""Test if DB construction works."""

import biom
import micom
import os.path as path
import pandas as pd
import qiime2 as q2
import q2_micom as q2m

this_dir = q2m.tests.this_dir

table = q2.Artifact.load(path.join(this_dir, "data", "table.qza")).view(
    biom.Table
)
taxa = q2.Artifact.load(path.join(this_dir, "data", "taxa.qza")).view(
    pd.Series
)
models = q2.Artifact.load(path.join(this_dir, "data", "species_models.qza"))


def test_build():
    d = q2m.build(
        table,
        taxa,
        models.view(q2m._formats_and_types.JSONDirectory),
        1,
        0.01,
        False
    )
    manifest = d.manifest.view(pd.DataFrame)
    print(manifest.T)
    assert manifest.shape[0] == 3
    for sa in manifest.sample_id.unique():
        com = micom.load_pickle(d.model_files.path_maker(model_id=sa))
        assert isinstance(com, micom.Community)
        assert len(com.reactions) > 3 * 95
        assert len(com.abundances) == 3



"""Test if DB construction works."""

import micom
import numpy as np
import os.path as path
import pandas as pd
import qiime2 as q2
import q2_micom as q2m

this_dir = q2m.tests.this_dir

meta = pd.read_csv(path.join(this_dir, "data", "test_models.tsv"), sep="\t")
meta.file = [path.join(this_dir, "data", f) for f in meta.file]
meta.set_index("id", inplace=True)
meta = q2.Metadata(meta)


def test_species():
    models = q2m.db(meta, "species", 1)
    files = models.manifest.view(pd.DataFrame)
    assert files.shape[0] == 3
    assert files.file[0] == "Escherichia_coli_1.json"
    cobra_model = micom.util.load_model(str(
        models.json_files.path_maker(model_id=files.id[0])))
    assert len(cobra_model.reactions) == 95
    assert np.allclose(
        cobra_model.optimize().objective_value,
        0.874,
        rtol=1e-3, atol=1e-3)


def test_genus():
    models = q2m.db(meta, "genus", 1)
    files = models.manifest.view(pd.DataFrame)
    assert files.shape[0] == 1
    assert files.file[0] == "Escherichia.json"
    cobra_model = micom.util.load_model(str(
        models.json_files.path_maker(model_id=files.id[0])))
    # one additional reaction due to averaged biomass
    assert len(cobra_model.reactions) == 96
    assert np.allclose(
        cobra_model.optimize().objective_value,
        0.874,
        rtol=1e-3, atol=1e-3)




"""Setup file for q2-micom."""
from setuptools import setup, find_packages

with open('README.md') as f:
    long_description = f.read()

setup(
    name="q2-micom",
    version="0.10.0",
    packages=find_packages(),
    package_data={"q2_micom": ["citations.bib", "assets/templates/*.html"]},
    author="Christian Diener",
    author_email="cdiener@isbscience.org",
    description="Plugin for metabolic modeling of microbial communities.",
    long_description=long_description,
    long_description_content_type='text/markdown',
    license="Apache License 2.0",
    url="https://github.com/micom-dev/q2-micom",
    classifiers=[
        "Development Status :: 4 - Beta",
        "Intended Audience :: Science/Research",
        "Topic :: Scientific/Engineering :: Bio-Informatics",
        "License :: OSI Approved :: Apache Software License",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
    ],
    # What does your project relate to?
    keywords="microbiome modeling metabolism community",
    entry_points={
        'qiime2.plugins': ['q2-micom=q2_micom.plugin_setup:plugin']
    },
    zip_safe=False,
    install_requires=[
        "cobra>=0.18.1",
        "pandas>=1.0",
        "loguru>=0.3.2",
        "micom==0.18.7",
        "jinja2>=2.10.3",
        "qiime2>=2020.6.0"
    ],
    python_requires=">=3.6",
    extras_require={
        "dev": [""],
        "test": ["coverage", "pytest", "pytest-cov", "flake8"],
    },
)



"""Test if tradeoff analysis works."""

import numpy as np
import os.path as path
import pandas as pd
import pytest
import qiime2 as q2
import q2_micom as q2m

this_dir = q2m.tests.this_dir

medium = q2.Artifact.load(path.join(this_dir, "data", "medium.qza")).view(
    pd.DataFrame
)
models = q2.Artifact.load(path.join(this_dir, "data", "build.qza"))

res = q2m.tradeoff(
    models.view(q2m._formats_and_types.CommunityModelDirectory),
    medium)


def test_tradeoff_values():
    assert np.allclose(res.tradeoff.min(), 0.1)
    assert np.allclose(res.tradeoff.max(), 1.0)
    assert res.tradeoff.nunique() == 10


def test_growth_rates():
    low = res.growth_rate[res.tradeoff == 0.1]
    high = res.growth_rate[res.tradeoff == 1.0]
    assert(low.sum() < high.sum())


def test_sane_tradeoff():
    with pytest.raises(ValueError):
        q2m.tradeoff(
            models.view(q2m._formats_and_types.CommunityModelDirectory),
            medium,
            tradeoff_min=0.5,
            tradeoff_max=0.4
        )
